# Copyright 2023 Agnostiq Inc.

import os
from functools import partial
from pathlib import Path
from typing import Dict

import toml
import yaml

from ..shared.classes.settings import DataRobotRegion, Settings


class AuthConfigManager:
    @staticmethod
    def get_config_file(settings: Settings = None) -> str:
        """
        Returns the path to the config file.

        Args:
            settings: The settings object to use. If None, the default settings will be used.

        Returns:
            The path to the config file.

        """

        if settings is None:
            settings = Settings()

        path = Path(settings.auth.config_file)
        path.parent.mkdir(parents=True, exist_ok=True)
        return str(path.resolve())

    @staticmethod
    def get_dr_config_file(settings: Settings = None) -> str:
        """
        Returns the path to the DataRobot config file. The 'DATAROBOT_CONFIG_FILE'
        environment variable takes precedence over the default config file path.

        Args:
            settings: The settings object to use. If None, the default settings will be used.

        Returns:
            The path to the DataRobot config file.

        """

        if settings is None:
            settings = Settings()

        # env override
        if dr_config_file := os.getenv("DATAROBOT_CONFIG_FILE"):
            path = Path(dr_config_file).expanduser().resolve()
            return str(path)

        dr_config_file = settings.auth.dr_config_file
        path = Path(settings.auth.dr_config_file).expanduser().resolve()
        path.parent.mkdir(parents=True, exist_ok=True)

        return str(path)

    @staticmethod
    def get_auth_request_headers(settings: Settings = None) -> Dict[str, str]:
        """
        Returns the headers to use for an authentication request.

        Args:
            settings: The settings object to use. If None, the default settings will be used.

        Returns:
            Authentication request headers as a dictionary.

        """

        if settings is None:
            settings = Settings()

        headers = {}

        if dr_api_token := AuthConfigManager.get_dr_api_token(settings):
            headers["Authorization"] = f"Bearer {dr_api_token}"

            dr_region = os.getenv("DATAROBOT_REGION", "")
            headers["x-dr-region"] = DataRobotRegion.validate(dr_region)

        else: 
            headers["x-api-key"] = AuthConfigManager.get_api_key(settings)

        return headers

    @staticmethod
    def save_token(token: str, settings: Settings = None) -> None:
        """
        Saves the authentication token to the config file.

        Args:
            token: The authentication token to save.
            settings: The settings object to use. If None, the default settings will be used.

        Returns:
            None

        """

        if settings is None:
            settings = Settings()

        auth_section_header = settings.auth.config_file_section
        token_keyname = settings.auth.config_file_token_keyname

        toml_dict = {auth_section_header: {}}
        toml_dict[auth_section_header][token_keyname] = token or ""

        with open(AuthConfigManager.get_config_file(settings), "w") as f:
            toml.dump(toml_dict, f)

    @staticmethod
    def get_token(settings: Settings = None) -> str:
        """
        Returns the authentication token from the config file.

        Args:
            settings: The settings object to use. If None, the default settings will be used.

        Returns:
            The authentication token.

        """

        if settings is None:
            settings = Settings()

        token = settings.auth.token
        if not token:
            auth_section_header = settings.auth.config_file_section
            token_keyname = settings.auth.config_file_token_keyname

            with open(AuthConfigManager.get_config_file(settings), "r") as f:
                toml_string = f.read()
                parsed_toml = toml.loads(toml_string)
                token = parsed_toml[auth_section_header][token_keyname]
        return token

    @staticmethod
    def get_dr_api_token(settings: Settings = None) -> str:
        """
        Returns the DataRobot API token from the DataRobot config file or environment variables,
        with the latter taking precedence over the former.

        The 'DATAROBOT_API_TOKEN' environment variable is checked first. If it is not set, the
        DataRobot config file is checked for the 'token' key. The config file path is
        determined by the 'DATAROBOT_CONFIG_FILE' environment variable, if set, else the
        default location.

        An empty string is returned if the environment variable is not set and the config file
        does not exist or does not contain the token.

        Args:
            settings: The settings object to use. If None, the default settings will be used.

        Returns:
            The DataRobot API token or an empty string if not found.

        """

        if settings is None:
            settings = Settings()

        # env override
        if dr_api_token := os.getenv("DATAROBOT_API_TOKEN"):
            return dr_api_token

        # settings override
        if dr_api_token := settings.auth.dr_api_token:
            return dr_api_token

        # config file
        dr_config_file = AuthConfigManager.get_dr_config_file(settings)
        if Path(dr_config_file).is_file():
            with open(dr_config_file, "r", encoding="utf-8") as f:
                return yaml.load(f, Loader=yaml.SafeLoader).get("token") or ""

        return ""

    @staticmethod
    def save_dr_api_token(dr_api_token: str, settings: Settings = None) -> None:
        """
        Saves the DataRobot API token to the DataRobot config file.

        Args:
            dr_api_token: The DataRobot API token to save.
            settings: The settings object to use. If None, the default settings will be used.

        Returns:
            None

        """

        if settings is None:
            settings = Settings()

        dr_config_file = AuthConfigManager.get_dr_config_file(settings)

        config: dict = {}
        if Path(dr_config_file).is_file():
            with open(dr_config_file, "r", encoding="utf-8") as f:
                config = yaml.load(f, Loader=yaml.SafeLoader) or {}

        config.update(token=str(dr_api_token))

        with open(dr_config_file, "w", encoding="utf-8") as f:
            yaml.dump(config, f, Dumper=yaml.SafeDumper)

    @staticmethod
    def get_api_key(settings: Settings = None) -> str:
        """
        Returns the API key from the config file.

        Args:
            settings: The settings object to use. If None, the default settings will be used.

        Returns:
            The API key.

        """

        if settings is None:
            settings = Settings()

        api_key = settings.auth.api_key
        if not api_key:
            auth_section_header = settings.auth.config_file_section
            api_keyname = settings.auth.config_file_api_key_keyname

            with open(AuthConfigManager.get_config_file(settings), "r") as f:
                toml_string = f.read()
                parsed_toml = toml.loads(toml_string)
                api_key = parsed_toml[auth_section_header][api_keyname]
        return api_key

    @staticmethod
    def save_api_key(api_key: str, settings: Settings = None) -> None:
        """
        Saves the API key to the config file.

        Args:
            api_key: The API key to save.
            settings: The settings object to use. If None, the default settings will be used.

        Returns:
            None

        """

        if settings is None:
            settings = Settings()

        auth_section_header = settings.auth.config_file_section

        api_keyname = settings.auth.config_file_api_key_keyname

        toml_dict = {auth_section_header: {}}
        toml_dict[auth_section_header][api_keyname] = api_key or ""

        with open(AuthConfigManager.get_config_file(settings), "w") as f:
            toml.dump(toml_dict, f)


get_token = partial(AuthConfigManager.get_token)
save_token = partial(AuthConfigManager.save_token)

get_api_key = partial(AuthConfigManager.get_api_key)
save_api_key = partial(AuthConfigManager.save_api_key)

get_dr_api_token = partial(AuthConfigManager.get_dr_api_token)
save_dr_api_token = partial(AuthConfigManager.save_dr_api_token)



# Copyright 2023 Agnostiq Inc.


from typing import Optional

from covalent_cloud.service_account_interface.client import get_client
from covalent_cloud.shared.classes.exceptions import handle_error
from covalent_cloud.shared.classes.settings import Settings, settings
from covalent_cloud.shared.schemas.volume import BaseVolume, Volume


def volume(
    name: str, vtype: Optional[str] = "OBJECT_STORAGE", settings: Optional[Settings] = settings
) -> Volume:
    """
    Creates or retrieves a persistent storage volume in Covalent Cloud. This function is used to manage volumes that provide persistent storage to workflow tasks, allowing them to read and write data across multiple workflow executions.

    The volumes are user-specific and remain available indefinitely until explicitly deleted. They are especially useful for workflows that require access to large datasets or need to maintain state between different executions.

    Args:
        name (str): The name of the volume. This name is used to create a new volume or retrieve an existing one.
        settings (Settings, optional): Configuration settings for connecting to the Covalent Cloud. Defaults to None, which uses the default settings.

    Returns:
        Volume: An object representing the created or retrieved volume.

    Examples:
        .. highlight:: python
        .. code-block:: python

            import covalent_cloud as cc
            import covalent as ct
            from covalent_cloud import volume

            # Creating a new volume
            my_volume = cc.volume("/mydata") # This will create a new volume named "/mydata" if it does not already exist.

            # Using the volume in a workflow

            # Define the workflow (here it is a single task workflow)
            @ct.lattice
            @ct.electron
            def my_workflow(x, y):
                # Workflow logic here to use "./mydata" volume
                pass

            # Dispatch the workflow with the volume
            run_id = cc.dispatch(my_workflow, volume=my_volume)(x=1, y=2)
            # The workflow now has access to the "/mydata" volume for persistent storage.
    """

    try:
        # will check if name is valid else throw error
        volume = BaseVolume(name=name)

        # should be valid name without any slashes since the validator cleaned it
        volume_name = volume.name

        client = get_client(settings)
        response = client.post(
            "/api/v2/volumes", request_options={"json": {"name": volume_name, "vtype": vtype}}
        )

        data = response.json()
        volume = Volume(**data)
        return volume
    except Exception as e:
        handle_error(e)



# Copyright 2024 Agnostiq Inc.

from pydantic import ConfigDict
from pydantic.dataclasses import dataclass

from covalent_cloud import CloudExecutor

from .models.gpu import GPU_TYPE

executor_plugin_name = "oci_cloud"


@dataclass(config=ConfigDict(extra="forbid"))
class OCICloudExecutor(CloudExecutor):
    """
    OCICloudExecutor is used to target BYOC infrastructure on OCI.
    """

    shape: str = "VM.Standard1.1"

    def __post_init__(self):
        """
        Post initialization method for the OCICloudExecutor.
        """
        super().__post_init__()

        num_units = int(self.shape.split(".")[-1])

        if "GPU" in self.shape:
            self.num_gpus = num_units
            self.gpu_type = str(GPU_TYPE.REMOTELY_SET)
        else:
            self.num_cpus = num_units

    @property
    def short_name(self) -> str:
        """
        Property which returns the short name
        of the executor used by Covalent for identification.

        Args:
            None

        Returns:
            The short name of the executor

        """
        return executor_plugin_name



# Copyright 2024 Agnostiq Inc.

import time
from enum import Enum
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from covalent_cloud.function_serve.models import Deployment

__all__ = [
    "DEPLOY_ELECTRON_PREFIX",
    "ACTIVE_DEPLOYMENT_POLL_INTERVAL",
    "SupportedMethods",
    "ServiceStatus",
    "rename",
    "wait_for_deployment_to_be_active",
]


DEPLOY_ELECTRON_PREFIX = "#__deploy_electron__#"
ACTIVE_DEPLOYMENT_POLL_INTERVAL = 10  # seconds


class SupportedMethods(str, Enum):
    """Supported HTTP methods for a function service."""

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"


class ServiceStatus(str, Enum):
    """Possible statuses for a function service."""

    NEW_OBJECT = "NEW_OBJECT"
    CREATING = "CREATING"
    ACTIVE = "ACTIVE"
    INACTIVE = "INACTIVE"
    ERROR = "ERROR"


WAITING_STATUSES = {ServiceStatus.NEW_OBJECT, ServiceStatus.CREATING}


class ServeAssetType(str, Enum):
    """Possible types for the ServeAsset `type` field."""

    ASSET = "Asset"
    JSON = "JSON"


def rename(name):
    def decorator(fn):
        fn.__name__ = name
        return fn

    return decorator


def wait_for_deployment_to_be_active(
    deployment: "Deployment",
    verbose: bool = False,
    wait_time_max: Union[int, float] = 3600,
) -> "Deployment":
    """Repeatedly reloads the deployment and waits for it to become active."""

    start_time = time.monotonic()
    while deployment.status in WAITING_STATUSES:
        if time.monotonic() - start_time >= wait_time_max:
            raise TimeoutError(
                f"Timed out after {wait_time_max / 60:.1f} minutes while waiting for "
                "the deployment to become active"
            )

        if verbose:
            print(f"Deployment {deployment.function_id} status: {deployment.status}")

        # Sleep for poll interval before checking again
        time.sleep(ACTIVE_DEPLOYMENT_POLL_INTERVAL)
        deployment.reload()

    # Reload anyways in case while loop is skipped
    deployment.reload()

    return deployment



# Copyright 2023 Agnostiq Inc.

"""Models for volumes"""
import re
from pathlib import Path
from typing import Optional

from pydantic import BaseModel, field_validator

from covalent_cloud.shared.classes.exceptions import CovalentSDKError

# Regex pattern for a valid directory name
# ^(/)? - Optionally starts with '/'
# ([\w-]+)$ - Followed by word characters or hyphens, and must end here
VOLUME_NAME_PATTERN = r"^/?([\w-]+)$"


class BaseVolume(BaseModel):
    name: str

    @staticmethod
    def get_valid_volume_name(name):
        """Get valid volume name.

        Args:
            name: Volume name

        Returns:
            Sanitized Volume name [str] if valid, else None

        """
        VOLUME_NAME_PATTERN = r"^/?([\w-]+)$"
        match = re.match(VOLUME_NAME_PATTERN, name)
        sanitized_name = match.group(1) if match else None
        is_matched = match is not None
        is_valid = is_matched and len(name) < 50
        return sanitized_name if is_valid else None

    @field_validator("name")
    def name_must_be_valid_directory(cls, v):
        valid_name = BaseVolume.get_valid_volume_name(v)
        if not valid_name:
            raise CovalentSDKError(
                "Name must be a valid directory name without subdirectories", "volume/invalid-name"
            )
        return valid_name

    @property
    def path(self):
        return f"/{self.name}"

    def __str__(self):
        return self.path

    def __truediv__(self, other):
        return Path(f"/volumes/{self.path}") / other


class Volume(BaseVolume):
    """Volume model"""

    id: int
    user_id: str
    efs_access_point_id: Optional[str] = None
    name: str
    deployment_id: Optional[str] = None
    bucket_prefix: Optional[str] = None



# Copyright 2023 Agnostiq Inc.

from unittest.mock import MagicMock

from covalent_cloud.shared.classes.api import APIClient
from covalent_cloud.shared.classes.settings import AuthSettings, Settings

MOCK_CC_API_KEY = "mock-cc-api-key"  # pragma: allowlist secret
MOCK_DR_API_TOKEN = "mock-dr-api-token"  # pragma: allowlist secret

MOCK_SETTINGS = Settings(
    auth=AuthSettings(
        dr_api_token=MOCK_DR_API_TOKEN,
    )
)

MOCK_LEGACY_SETTINGS = Settings(
    auth=AuthSettings(
        api_key=MOCK_CC_API_KEY,
    )
)

MOCK_GLOBAL_HEADERS = {
    "Authorization": f"Bearer {MOCK_DR_API_TOKEN}",
    'x-dr-region': '',
}

MOCK_LEGACY_GLOBAL_HEADERS = {
    "x-api-key": MOCK_CC_API_KEY,
}
MOCK_REQUEST_OPTIONS = {
    "params": {"a": 1},
}


class TestAPIClient:

    MOCK_URI = "https://www.example.com"

    def test_get_global_headers(self):
        # substituting token avoids fetching/creating from config file
        api = APIClient(host_uri=self.MOCK_URI, headers={"CLIENT": "TEST"}, settings=MOCK_SETTINGS)
        assert api.get_global_headers() == {"CLIENT": "TEST", **MOCK_GLOBAL_HEADERS}

    def test_get_legacy_global_headers(self):
        # substituting token avoids fetching/creating from config file
        api = APIClient(host_uri=self.MOCK_URI, headers={"CLIENT": "TEST"}, settings=MOCK_LEGACY_SETTINGS)
        assert api.get_global_headers() == {"CLIENT": "TEST", **MOCK_LEGACY_GLOBAL_HEADERS}

    def test_prepare_request(self):
        # substituting token avoids fetching/creating from config file
        api = APIClient(host_uri="http://some.site/", headers={}, settings=MOCK_SETTINGS)
        uri, options = api.prepare_request("/test", {"data": 123})
        print("OPTIONS:", options)
        assert uri == "http://some.site/test"
        assert options == {"headers": MOCK_GLOBAL_HEADERS, "data": 123}

    def test_prepare_request_with_legacy_headers(self):
        # substituting token avoids fetching/creating from config file
        api = APIClient(host_uri="http://some.site/", headers={}, settings=MOCK_LEGACY_SETTINGS)
        uri, options = api.prepare_request("/test", {"data": 123})
        print("OPTIONS:", options)
        assert uri == "http://some.site/test"
        assert options == {"headers": MOCK_LEGACY_GLOBAL_HEADERS, "data": 123}

    def test_get_request_options(self):
        # substituting token avoids fetching/creating from config file
        api = APIClient(host_uri=self.MOCK_URI, headers={}, settings=MOCK_SETTINGS)
        options = api.get_request_options({"body": {"a": 1}})
        assert options["body"] == {"a": 1}

    def test_post(self, mocker):

        api_mock = APIClient(host_uri=self.MOCK_URI)
        requests_mock = MagicMock()
        mocker.patch(
            "covalent_cloud.shared.classes.api.requests.Session.__enter__",
            return_value=requests_mock,
        )

        mock_request_options = {"json": {"a": 1}, "data": b"ef8"}

        mocker.patch.object(api_mock, "get_request_options", return_value=mock_request_options)
        api_mock.post("/api/v0/resource", mock_request_options)
        requests_mock.post.assert_called_with(
            f"{self.MOCK_URI}/api/v0/resource", **mock_request_options
        )

    def test_get(self, mocker):

        api_mock = APIClient(host_uri=self.MOCK_URI)
        requests_mock = MagicMock()
        mocker.patch(
            "covalent_cloud.shared.classes.api.requests.Session.__enter__",
            return_value=requests_mock,
        )
        mocker.patch.object(api_mock, "get_request_options", return_value=MOCK_REQUEST_OPTIONS)
        api_mock.get("/api/v0/resource", MOCK_REQUEST_OPTIONS)
        requests_mock.get.assert_called_with(
            f"{self.MOCK_URI}/api/v0/resource", **MOCK_REQUEST_OPTIONS
        )

    def test_delete(self, mocker):

        api_mock = APIClient(host_uri=self.MOCK_URI)
        requests_mock = MagicMock()
        mocker.patch(
            "covalent_cloud.shared.classes.api.requests.Session.__enter__",
            return_value=requests_mock,
        )
        mocker.patch.object(api_mock, "get_request_options")
        api_mock.delete("/api/v0/resource")
        requests_mock.delete.assert_called_with(f"{self.MOCK_URI}/api/v0/resource")

    def test_put(self, mocker):

        api_mock = APIClient(host_uri=self.MOCK_URI)
        requests_mock = MagicMock()
        mocker.patch(
            "covalent_cloud.shared.classes.api.requests.Session.__enter__",
            return_value=requests_mock,
        )
        mocker.patch.object(api_mock, "get_request_options", return_value=MOCK_REQUEST_OPTIONS)
        api_mock.put("/api/v0/resource", MOCK_REQUEST_OPTIONS)
        requests_mock.put.assert_called_with(
            f"{self.MOCK_URI}/api/v0/resource", **MOCK_REQUEST_OPTIONS
        )



# Copyright 2024 Agnostiq Inc.

import json
import warnings
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Type, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, field_validator
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from covalent import TransportableObject
from covalent_cloud.cloud_executor.cloud_executor import CloudExecutor
from covalent_cloud.cloud_executor.oci_cloud_executor import OCICloudExecutor
from covalent_cloud.function_serve.client import FunctionServeClient
from covalent_cloud.function_serve.common import ServeAssetType, ServiceStatus, SupportedMethods
from covalent_cloud.service_account_interface.client import get_deployment_client
from covalent_cloud.shared.schemas.volume import Volume

__all__ = [
    "Endpoint",
    "FunctionServiceModel",
    "Deployment",
]

COVALENT_CUSTOM_ENCODERS = {
    UUID: str,
    datetime: lambda dt: dt.isoformat(),
    timedelta: lambda td: td.total_seconds(),
}


# Type alias to represent a serialized object type
SerializedObjectType = bytes


class CustomBaseModel(BaseModel):
    """
    Custom Pydantic base model that converts certain data types to strings for json serialization.

    This is useful for when we want to serialize a Pydantic model to json, but the model contains data types that are not json serializable.

    Example:
        class MyModel(CustomBaseModel):
            id: uuid.UUID
            now: datetime.datetime

        model = MyModel(id=uuid.uuid4(), now=datetime.utcnow())
        json_str = model.json()  # was json serializable where before is raised a TypeError
        dict = model.model_dump() # dict is json serializable
    """

    @classmethod
    def encode_value(cls, value: Any, encoders: Dict[Type[Any], Callable[[Any], Any]]) -> Any:
        if isinstance(value, BaseModel):
            return value.model_dump(using_encoder=True, encoders=encoders)
        elif isinstance(value, list):
            return [cls.encode_value(item, encoders) for item in value]
        elif isinstance(value, dict):
            return {k: cls.encode_value(v, encoders) for k, v in value.items()}
        else:
            for encoder_type, encoder_func in encoders.items():
                if isinstance(value, encoder_type):
                    return encoder_func(value)
            return value

    def model_dump(
        self,
        *,
        using_encoder: bool = False,
        encoders: Dict[Type[Any], Callable[[Any], Any]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        if not using_encoder:
            encoders = COVALENT_CUSTOM_ENCODERS
        encoded_dict = super().model_dump(**kwargs)
        return {key: self.encode_value(value, encoders) for key, value in encoded_dict.items()}

    def json(self, **kwargs) -> str:
        return json.dumps(self.model_dump(**kwargs))

    def dict(self, **kwargs) -> Dict[str, Any]:
        return self.model_dump(**kwargs)

    class Config:
        json_encoders = COVALENT_CUSTOM_ENCODERS


class ServeAsset(BaseModel):
    type: Union[str, ServeAssetType] = ServeAssetType.ASSET
    id: Optional[str] = None
    url: Optional[str] = None
    serialized_object: SerializedObjectType

    def upload(self):
        from covalent_cloud.function_serve.assets import AssetsMediator

        return AssetsMediator.upload_asset(self)

    @field_validator("type")
    @classmethod
    def type_as_str(cls, type_: Union[str, ServeAssetType]):
        if isinstance(type_, Enum):
            return type_.value
        return type_


class Endpoint(BaseModel):  # TODO after v0: include executor and resources
    """
    Information about an individual endpoint for a function service.
    """

    method: str = SupportedMethods.POST
    """HTTP method for the endpoint; 'GET', 'POST', 'PUT', etc."""

    route: str
    """Route for the endpoint; e.g. '/predict'."""

    name: str
    """Name of the endpoint"""

    endpoint_fn: ServeAsset
    """Serialized function that implements the endpoint."""

    endpoint_fn_source: str
    """Source code of the function that implements the endpoint."""

    streaming: bool
    """"Indicates if the endpoint can stream its response."""

    description: str
    """Description of the endpoint."""

    test_endpoint_enabled: bool
    """Flag that indicates one-click testing is possible for the endpoint."""

    def __call__(self, *args, **kwargs):
        return TransportableObject.deserialize(self.endpoint_fn.serialized_object).get_deserialized()(*args, **kwargs)


class FunctionServiceModel(CustomBaseModel):
    """
    Encapsulates all information relevant to a function service.
    """

    model_config = ConfigDict(extra="forbid")
    """Pydantic configuration for the model."""

    title: Optional[str] = ""
    """Title of the function service."""

    auth: Optional[bool] = None
    """Flag that indicates if the function service endpoints require authorization."""

    description: Optional[str] = ""
    """User specified description of the function service."""

    executor: Union[CloudExecutor, OCICloudExecutor]
    """Executor that defines the total resources (NOTE: revise after v0!)"""

    compute_share: Optional[float] = 1.0
    """Share of the total compute resources to be allocated to the function service. (NOTE: revise after v0!)"""

    tags: Optional[Union[str, List[str]]] = ""
    """List of optional tags for the function service."""

    init_fn: ServeAsset
    """Serialized service initializer function."""

    init_fn_args: ServeAsset
    """Serialized arguments to the service initializer function."""

    init_fn_args_json: ServeAsset
    """JSON array containing the initializer function arguments."""

    init_fn_kwargs: ServeAsset
    """Serialized keyword arguments to the service initializer function."""

    init_fn_kwargs_json: ServeAsset
    """JSON object containing the initializer function keyword arguments."""

    init_fn_source: str
    """Source code of the service initializer function."""

    endpoints: List[Endpoint]
    """List of endpoints for the function service."""

    volume: Optional[Volume] = None
    """Volume to be mounted to the function service."""

    root_dispatch_id: Optional[str] = None
    """Root dispatch ID of the function service, only valid for workflow-based deployments."""

    @field_validator("compute_share")
    @classmethod
    def compute_share(cls, compute_share: float):
        """Validate the compute shares for the function service."""
        if compute_share <= 0 or compute_share > 1:
            raise ValueError("compute_shares must be a float of value > 0 and <= 1.0.")
        return compute_share


class RouteInfo(BaseModel):
    """
    Contains information about a specific service route.
    This model is a partial `Endpoint` model.
    """

    route: str
    """Route for the endpoint; e.g. '/predict'."""

    method: SupportedMethods
    """HTTP method for the endpoint; 'GET', 'POST', 'PUT', etc."""

    streaming: bool
    """"Indicates if the endpoint can stream its response."""

    description: Optional[str] = ""
    """Description of the endpoint."""


class Deployment(BaseModel):
    """
    Returned by `cc.deploy` function and when `@service` decorated function is
    called within a workflow.
    """

    function_id: str
    """Unique identifier for the deployment."""

    address: Optional[str] = None
    """Address of the deployment usable by the user."""

    name: str
    """Name of the deployment."""

    description: str
    """Description of the deployment."""

    routes: List[RouteInfo]
    """All the registered routes for the deployment."""

    status: ServiceStatus
    """Status of the deployment."""

    tags: List[str] = Field(default_factory=list)
    """List of optional tags associated with the deployment."""

    token: Optional[str] = None
    """Authorization token for requests to the endpoints."""

    auth: Optional[bool] = None
    """Flag that indicates if the deployment endpoints require authorization."""

    error: Optional[str] = None
    """Error message associated with creating the deployment."""

    stderr: Optional[str] = None
    """Error message associated with initializing the deployment."""

    # Making this a static method since we don't need to access the class.
    @staticmethod
    def from_function_record(data: Dict[str, Any]) -> "Deployment":
        """
        Create a Deployment object from a function's database record.
        """

        return Deployment(
            function_id=data["id"],
            description=data.get("description"),
            address=str(data["invoke_url"]),
            name=str(data["title"]),
            routes=[RouteInfo(**ep) for ep in data["endpoints"]],
            status=data["status"],
            tags=data["tags"] or [],
            token=data.get("inference_keys", [])[0].get("key")
            if data.get("inference_keys")
            else None,
            auth=data["auth"],
            error=data.get("error"),
            stderr=data.get("stderr"),
        )

    # Not making this a model validator since in that case
    # it will try to attach the method to this model
    # and then Pydantic will raise an error since it cannot
    # serialize a 'method' type object.
    def attach_route_methods(self, overwrite: bool = False) -> "Deployment":
        """
        Attach methods to the Deployment object for each route.
        """

        fs_client = FunctionServeClient(
            function_id=self.function_id,
            host=self.address,
            token_getter=lambda: self.token,
        )

        # Built-in methods.
        setattr(self, "info", fs_client.info)
        setattr(self, "teardown", fs_client.teardown)

        # User-defined methods.
        for route_info in self.routes:

            # Check for name collisions.
            _name = route_info.route.lstrip("/").replace("-", "_")
            if hasattr(self, _name) and not overwrite:
                warnings.warn(
                    f"Object already has an attribute '{_name}'. Skipping method attachment. "
                    f"Please make requests to '{self.address}{route_info.route}' manually.",
                    category=UserWarning,
                )
                continue

            # Sync version.
            _new_sync_method = fs_client.make_request_method(
                route=route_info.route,
                method=route_info.method,
                streaming=route_info.streaming,
                is_async=False,
            )
            _new_sync_method.__name__ = _name
            _new_sync_method.__doc__ = route_info.description
            setattr(self, _name, _new_sync_method)

            # Async version. Append prefix to name and description.
            _async_name = f"async_{_name}"
            _new_async_method = fs_client.make_request_method(
                route=route_info.route,
                method=route_info.method,
                streaming=route_info.streaming,
                is_async=True,
            )
            _new_async_method.__name__ = _async_name
            _new_async_method.__doc__ = "(async) " + route_info.description
            setattr(self, _async_name, _new_async_method)

        return self

    def reload(self) -> None:
        """
        Reload the deployment info with updated information
        such as most recent status or address of the deployment.
        """

        if not self.function_id:
            raise ValueError("Function ID is not set.")

        deployment_client = get_deployment_client()

        # Reload the deployment info.
        response = deployment_client.get(f"/functions/{self.function_id}")

        new_deployment = Deployment.from_function_record(response.json())
        self.__dict__.update(new_deployment.__dict__)

        # Attach route methods for ease of use
        self.attach_route_methods(overwrite=True)

    def _stringify_routes(self) -> Table:
        """Get rich Table representation of route info."""

        routes_table = Table(box=None, show_header=False, padding=(0, 1))
        for i, route in enumerate(self.routes):
            routes_table.add_row("Route", f"{route.method.value} {route.route}")
            routes_table.add_row("Streaming", "Yes" if route.streaming else "No")
            routes_table.add_row("Description", route.description)
            if i < len(self.routes) - 1:
                routes_table.add_row("", "")  # empty row for spacing

        return routes_table

    def __str__(self) -> str:
        console = Console()
        info_table = Table(box=None, show_header=False)
        info_table.add_row("Name", self.name)
        info_table.add_row("Description", self.description)
        info_table.add_row("Function ID", self.function_id)
        info_table.add_row("Address", self.address)
        info_table.add_row("Status", self.status.value)

        if len(self.tags) > 0:
            info_table.add_row("Tags", ", ".join(self.tags))

        info_table.add_row("Auth Enabled", "Yes" if self.auth else "No")

        _panel_width = 90

        # Create a Panel for the main info
        info_panel = Panel(info_table, title="Deployment Information", width=_panel_width)

        # Create panel for error messages
        # Give `stderr` priority over `error`
        error_string = (self.stderr or self.error or "").strip()
        error_panel = Panel(error_string, title="Exception", width=_panel_width)

        # Prepare route info as Tables inside Panels
        routes_panel = Panel(self._stringify_routes(), title="Endpoints", width=_panel_width)

        # Convert everything to string using Rich Console's capture feature
        with console.capture() as capture:

            console.print(info_panel)

            if error_string:
                console.print(error_panel, style="red")

            console.print(routes_panel)

            if self.token:
                console.print(f"Authorization token:\n{self.token}")

        return capture.get()

    class Config:
        """Model configuration."""

        # Need to allow extra for dynamic method assignment.
        extra = "allow"



# Copyright 2023 Agnostiq Inc.

import os
import re
from dataclasses import asdict
from datetime import timedelta
from typing import Dict, Optional, Union

import arrow
from pydantic import ConfigDict, Field, field_validator, model_validator
from pydantic.dataclasses import dataclass

from ..shared.classes.helpers import check_env_is_ready
from ..shared.classes.settings import Settings, settings
from .models.gpu import DEPRECATED_GPUS, GPU_TYPE

executor_plugin_name = "cloud"
unit_multipliers = {"GB": 1024, "MB": 1}
gpu_type_allowed_counts = {
    GPU_TYPE.A100: [1, 2, 4, 8],
    GPU_TYPE.H100: [1, 2, 4, 8],
    GPU_TYPE.V100: [1, 4, 8],
    GPU_TYPE.A10: [1, 4, 8],
    GPU_TYPE.A6000: [1, 2, 4, 8],
    GPU_TYPE.A4000: [1, 2, 4, 8, 10],
    GPU_TYPE.A5000: [1, 2, 4, 8],
    GPU_TYPE.T4: [1, 4],
    GPU_TYPE.L40: [1, 2, 4, 8],
}

EXTRAS_ALLOWED = os.environ.get("COVALENT_CLOUD_EXECUTOR_EXTRAS_ALLOWED", "0") == "1"


@dataclass(config=ConfigDict(extra="allow" if EXTRAS_ALLOWED else "forbid"))
class CloudExecutor:
    """
    CloudExecutor represents a configuration for executing a Covalent workflow on the Covalent Cloud.
    This class allows users to configure the resources (such as the number of CPUs, memory, GPUs, and GPU type) and the software environment for a Covalent workflow that will be executed on the Covalent Cloud. The time limit for the workflow execution can also be set.

    Attributes:

        num_cpus (int, optional): Number of CPUs to be used for the workflow. Defaults to 1.
        cpu_type (str, optional): Model of the CPU to be used for the workflow. Defaults to an empty string.
        memory (int, optional): Amount of memory (in MB) to be used for the workflow. Defaults to 1024.
        num_gpus (int, optional): Number of GPUs to be used for the workflow. Defaults to 0.
        gpu_type (Union[str, cloud_executor.GPU_TYPES], optional): Type of GPU to be used for the workflow. Defaults to an empty string.
        env (str, optional): Name of the software environment to be used for the workflow. Defaults to "default".
        time_limit (Union[int, timedelta], optional): Time limit for the workflow execution, in seconds or as a timedelta. Defaults to 1800s (30 mins). Alternatively can take human readable string in format 'in <number> <unit(s)>'
        volume_id (int, optional): ID of the volume to be used for the workflow. Defaults to None.
        settings (Settings, optional): Settings object to be used for the workflow. Defaults to the global settings object.
        validate_environment (bool, optional): Whether to validate the environment before dispatching the workflow. Defaults to True.
    Examples:

        .. highlight:: python
        .. code-block:: python

            # create a CloudExecutor with default resource configuration
            # executor = CloudExecutor()

            # create a custom CloudExecutor with specified resources and environment
            # executor = CloudExecutor(
            #     num_cpus=4,
            #     memory=2048,
            #     num_gpus=1,
            #     gpu_type="NVIDIA-Tesla-V100",
            #     env="my_custom_env",
            #     time_limit="in 30 minutes"  # 30 minutes
            # )

            import covalent as ct
            from covalent_cloud import CloudExecutor

            cloud_executor1 = CloudExecutor(num_cpus=1, memory=1024)
            cloud_executor2 = CloudExecutor(num_cpus=2, memory=2048)
            cloud_executor3 = CloudExecutor(num_cpus=1, memory=512)

            # Define manageable tasks as electrons with different cloud executors
            @ct.electron(executor=cloud_executor1)
            def add(x, y):
                return x + y

            @ct.electron(executor=cloud_executor2)
            def multiply(x, y):
                return x * y

            @ct.electron(executor=cloud_executor3)
            def divide(x, y):
                return x / y

            # Define the workflow as a lattice
            @ct.lattice
            def workflow(x, y):
                r1 = add(x, y)
                r2 = [multiply(r1, y) for _ in range(4)]
                r3 = [divide(x, value) for value in r2]
                return r3

            # Import the Covalent Cloud module
            import covalent_cloud as cc

            # Dispatch the workflow to the Covalent Cloud
            dispatch_id = cc.dispatch(workflow)(1, 2)
            result = cc.get_result(dispatch_id, wait=True)
            print(result)

    """

    num_cpus: int = 1
    cpu_type: str = ""
    memory: Union[int, str] = 1024
    num_gpus: int = 0
    gpu_type: Union[str, GPU_TYPE] = ""
    env: str = "default"
    time_limit: Union[int, timedelta, str] = 60 * 30
    volume_id: Optional[int] = None
    settings: Dict = Field(default_factory=settings.model_dump)
    validate_environment: bool = True

    def __post_init__(self, **kwargs):
        # check if the environment is valid
        if self.validate_environment:
            check_env_is_ready(self.env, Settings.model_validate(self.settings))

        # if gpu type is specified, but num_gpus is not, set num_gpus to 1
        if self.gpu_type and self.num_gpus == 0:
            self.num_gpus = 1

        # check if gpu type and num are consistent
        if (
            os.environ.get("COVALENT_CLOUD_DISABLE_GPU_VALIDATION") is None
            and (self.num_gpus > 0 or self.gpu_type)
            and GPU_TYPE(self.gpu_type) in gpu_type_allowed_counts
            and self.num_gpus not in gpu_type_allowed_counts.get(GPU_TYPE(self.gpu_type), [])
        ):
            raise ValueError(
                f"Invalid number of GPUs for {self.gpu_type}. "
                f"Please choose from: {gpu_type_allowed_counts.get(GPU_TYPE(self.gpu_type))}"
            )

        self.gpu_type = str(self.gpu_type)

        self._extras = {}
        if EXTRAS_ALLOWED:
            for k, v in self.__dict__.items():
                if k not in self.__dataclass_fields__ and k != "_extras":
                    self._extras[k] = v

    @property
    def short_name(self) -> str:
        """
        Property which returns the short name
        of the executor used by Covalent for identification.

        Args:
            None

        Returns:
            The short name of the executor

        """
        return executor_plugin_name

    # Validators:
    @model_validator(mode="after")
    def gpu_type_required(self) -> "CloudExecutor":
        """
        Validator which ensures that GPU type is specified if any GPUs are requested.

        Returns:
            The validated instance

        """
        if self.num_gpus > 0 and not self.gpu_type:
            raise ValueError("GPU type must be specified if num_gpus > 0")

        return self

    @field_validator("num_cpus", "memory", "time_limit")
    @classmethod
    def gt_than_zero(cls, v: int) -> int:
        """
        Validator which ensures that the value is greater than 0.

        Args:
            v: The value to validate

        Returns:
            The validated value

        """
        if v <= 0:
            raise ValueError(f"{v} must be greater than 0")

        return v

    @field_validator("gpu_type")
    @classmethod
    def gpu_types_non_deprecated(cls, v: str) -> str:
        """
        Validator which ensures that the GPU type is not deprecated.

        Args:
            v: The value to validate

        Returns:
            The validated value

        """
        if v and v in DEPRECATED_GPUS:
            raise ValueError(f"GPU type '{v}' is now deprecated.")
        return v

    @field_validator("memory", mode="before")
    @classmethod
    def memory_to_int(cls, v: Union[int, str]) -> int:
        """
        Validator which converts the memory value to an integer.

        Args:
            v: The value to validate

        Returns:
            The validated value

        """
        if isinstance(v, str):
            # grab the number and the unit
            match = re.match(r"(\d+)\s*([A-Za-z]+)", v)
            if match:
                memory_value, unit = match.groups()
            else:
                raise ValueError("Invalid memory string format")

            # convert to MB (int)
            try:
                memory_value = int(memory_value)
                unit = unit.strip()
                v = int(memory_value * unit_multipliers[unit])
            except ValueError as e:
                raise ValueError(f"Invalid memory value: {v}.") from e

        return v

    @field_validator("time_limit", mode="before")
    @classmethod
    def time_limit_to_int_seconds(cls, v: Union[int, timedelta, str]) -> int:
        """
        Validator which converts the time limit value to seconds.

        Args:
            v: The value to validate

        Returns:
            The validated value

        """
        if isinstance(v, timedelta):
            v = v.total_seconds()

        elif isinstance(v, str):
            v = _time_string_total_seconds(v)

        return int(v)

    def to_dict(self) -> dict:
        """
        Return a JSON-serializable dictionary representation of this object.

        Args:
            None

        Returns:
            The JSON-serializable dictionary representation of this object.

        """
        return {
            "type": str(self.__class__),
            "short_name": self.short_name,
            "attributes": asdict(self) | self._extras,
        }

    @classmethod
    def from_dict(cls, data: dict) -> "CloudExecutor":
        """
        Create a CloudExecutor object from a dictionary.

        Args:
            data: The dictionary to create the CloudExecutor object from.

        Returns:
            The CloudExecutor object created from the dictionary.

        """
        return cls(**data)


def _time_string_total_seconds(time_limit: str) -> float:
    """Convert a time limit string to seconds.

    Args:
        time_limit: A string specifying the time limit.

    Returns:
        Total seconds in the time limit.

    """
    time_limit = time_limit.strip()

    if not time_limit:
        raise ValueError("Empty time limit string.")

    if "-" in time_limit:  # ad hoc disallow any negatives
        raise ValueError("The '-' character is not allowed in time limit strings.")

    is_human = any(w in time_limit for w in ["hours", "minutes", "seconds"])

    if is_human and re.findall(r"(?:\d+:\d+:\d+\s|\s\d+:\d+:\d+)", time_limit):
        raise ValueError("Mixed time formats are not allowed.")

    now = arrow.utcnow()
    # parse the 'later' point in time
    try:
        if hs_ms_ss := re.match(r"(\d+):(\d+):(\d+)$", time_limit):  # e.g. '05:45:00'
            hs, ms, ss = hs_ms_ss.groups()
            later = now.shift(hours=int(hs), minutes=int(ms), seconds=int(ss))

        else:  # e.g. '5 hours and 45 minutes'
            time_limit = time_limit if time_limit.startswith("in ") else f"in {time_limit}"
            later = now.dehumanize(time_limit)

    except Exception as e:
        raise ValueError(
            f"Invalid time limit string: '{time_limit}'. "
            "Please provide a valid time limit, e.g. '5 hours and 45 minutes' or '05:45:00'."
        ) from e

    # amount of time is the difference between datetimes
    return (later - now).total_seconds()



# Copyright 2023 Agnostiq Inc.

"""Software environment management module."""

import re
import sys
import tempfile
import time
import warnings
from typing import Dict, List, Optional, Tuple, Union

import requests
import yaml
from packaging import version

from covalent_cloud import get_client
from covalent_cloud.shared.classes.helpers import check_env_is_ready
from covalent_cloud.shared.classes.settings import Settings, settings
from covalent_cloud.swe_management.models.environment import (
    BaseImageConfig,
    Environment,
    EnvironmentRuntimeConfig,
)

from ..shared.classes.exceptions import CovalentAPIKeyError, CovalentGenericAPIError

__all__ = [
    "create_env",
    "get_envs",
    "delete_env",
]

PY_VERSION_WARNING_TEMPLATE = (
    "Specified version '{dependency}' is either invalid or does not "
    "guarantee Python {local_version} to match your local environment. "
    "Dispatches and deployments may fail unexpectedly if the local version "
    "differs from the cloud environment. To avoid this warning, omit the "
    "'python' conda dependency or specify exactly the local Python version. "
    "Note: `create_env` selects your local Python version by default."
)


def get_pip_pkgs(pip: Union[str, List[str]]) -> List[str]:
    """
    Unpacks the pip packages in the requirements.txt file and combines it into a list of required pip packages.

    """

    if not isinstance(pip, list):
        pip = [pip]

    pip_pkgs = []
    for pkg in pip:
        if pkg.endswith(".txt"):
            with open(pkg, "r") as f:
                pip_pkgs += f.read().splitlines()
        else:
            pip_pkgs.append(pkg)

    return pip_pkgs


def unpack_conda_pkgs(
    conda: Union[str, List[str], Dict[str, List[str]]]
) -> Tuple[List[str], List[str]]:
    """
    Unpacks the conda packages in the environment.yml file and combines it into a dictionary of required conda packages.

    Returns:
        channels, dependencies: channels and dependencies according to the conda environment.yml file. Note that these terms are chosen according to the conda nomenclature.

    """

    channels, dependencies = [], []

    if isinstance(conda, dict):
        channels = conda.get("channels", [])
        dependencies = conda.get("dependencies", [])

    elif isinstance(conda, str):
        if conda.endswith(".yml"):
            with open(conda, "r") as f:
                parsed_conda_env_yaml = yaml.safe_load(f)

            channels = parsed_conda_env_yaml.get("channels", [])
            dependencies = parsed_conda_env_yaml.get("dependencies", [])

    elif isinstance(conda, list):
        dependencies = conda

    return channels, dependencies


def missing_dependency(
    dependency: str, dependencies: List[str], exact: bool = False
) -> Optional[str]:
    """
    Checks if the dependencies list contains anything starting with the string.

    Args:
        dependency: String to check if it's in the dependencies list.
        dependencies: List of dependencies to be installed in the environment.
        exact: If True, checks if the dependency is exactly the same as the one in the dependencies list.
        verbose: If True, returns the exact string for the dependency that is missing.

    Returns:
        The missing dependency as a string, if applicable, or None if no dependency is missing.
    """

    for dep in dependencies:
        # check if any dependency starts with the string
        if (not exact and str(dep).startswith(dependency)) or (
            exact and str(dep).strip() == dependency.strip()
        ):
            return str(dep).strip()

    return None


def _check_local_python_dependency(dependency: str) -> None:
    """
    Checks the local python version and warns the user if the requested python version is different from the local python version.
    Args:
        dependency: The python dependency (python version) to be checked against the local python installation.
    Returns:
        None

    """
    # Get local release tuple
    local_release = tuple(sys.version_info[:3])
    local_version = ".".join(map(str, local_release))

    # Get remote version or version range
    pattern = r"python(=|>=|<=|<|>|!=)(.*)"
    pattern_match = re.match(pattern, dependency)

    if pattern_match is None:
        raise ValueError(f"Invalid python dependency: {dependency}")

    warn = False
    try:
        # Separate pattern groups
        sep, remote_version = pattern_match.groups()

        # Try to parse the version code
        remote_release = None
        if "," in pattern_match.groups()[1]:
            # Try to parse anyway just in case there is a detectable error
            for r in remote_version.split(","):
                version.parse(r.strip("=><!"))
            warn = True
        else:
            remote_release = version.parse(remote_version).release

    except version.InvalidVersion as e:
        raise ValueError(f"Invalid python version: {remote_version}") from e

    # Finally, show warning if version is valid but not exactly local version
    if warn or remote_release != local_release or sep != "=":
        warnings.warn(
            PY_VERSION_WARNING_TEMPLATE.format(
                local_version=local_version,
                dependency=dependency,
            )
        )


def create_env(
    name: str,
    pip: Optional[Union[str, List[str]]] = [],
    conda: Optional[Union[str, List[str], Dict[str, List[str]]]] = [],
    settings: Optional[Settings] = settings,
    wait: Optional[bool] = False,
    timeout: Optional[int] = 1800,
    base_image: Optional[str] = None,
    nvidia: bool = False,
) -> None:
    """
    Sends the create request to the Covalent Cloud server with the environment dependency list.

    Args:
        name: Identifier/name for the software environment.

        pip: Python packages to be installed in the environment using pip. This value can be a string `requirements.txt` and/or a list of packages. Note, that if it's a list, it's possible that one of the values is the string `requirements.txt`. In case a `requirements.txt` is passed, it will be parsed into a list of packages and combined with the list of packages passed.`

        conda: List of packages to be installed in the environment using conda. This value can either be a list of packages, a filepath to `environment.yml`. It could also be a dictionary with channels and dependencies as keys and a list of strings as their values. For example:

            conda={
                        "channels": ["conda-forge", "defaults"],
                        "dependencies": ["numpy=1.21.*", "xarray=0.15.1"],
            }

        Whatever is passed, it will be parsed into a dictionary as shown above and sent as JSON to the Covalent Cloud server. If a list of packages is provided, they will be installed using the default conda channel.

        settings: Settings object with the dispatcher URI.

        wait: If True, waits until the environment is ready before returning.

        timeout: Timeout in seconds for the environment to be ready.

        base_image: Base image of the runtime environment. If base_image is provided, it should be of the form <url>/<image>:<tag> (e.g. docker.io/python:3.9.6), where the
        :tag is optional and will default to latest if not specified. Only publicly accessible images are supported for now.

    Returns:
        None

    Examples:
        Create an environment with a list of packages:
            >>> create_env("test-env", ["typing"], ["numpy=1.21.*", "xarray=0.15.1"])

        Create an environment with a filepath to `environment.yml`:
            >>> create_env("test-env", "requirements.txt", "environment.yml")

        Create an environment with a dictionary of channels and dependencies
            >>> create_env("test-env", "requirements.txt", {"channels": ["conda-forge", "defaults"], "dependencies": ["numpy=1.21.*", "xarray=0.15.1"]})

      Create an environment with a different base image:
            >>> create_env("test-env", ["typing"], ["numpy=1.21.*", "xarray=0.15.1"], base_image="docker.io/python:slim")

    Note:
        - In case of a conflict of package between pip and conda, pip will take precedence and the conda one will be ignored.
        - If the python dependency is not provided, the local environment python version will be used.
        - If the python dependency that is provided is different from the local environment python version the user will be give a warning.
        - If the pip dependency is not provided, it will be added to the dependencies.

    """
    pip = pip or []
    conda = conda or {}

    notes = []
    pip_pkgs = get_pip_pkgs(pip)
    channels, dependencies = unpack_conda_pkgs(conda)

    # check dependencies contains pip and if not add pip
    dependency = missing_dependency("pip", dependencies, exact=True)
    if not dependency:
        dependencies = ["pip"] + dependencies
        notes.append("pip was added to the dependencies.")

    # add pip packages to dependencies after checking if pip is explicitly in the dependencies
    dependencies.append({"pip": pip_pkgs})

    # check dependencies contains python and if not set to current python version
    dependency = missing_dependency("python", dependencies)
    if not dependency:
        python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
        dependencies = [f"python={python_version}"] + dependencies
        notes.append(f"Python version {python_version} was added to the dependencies.")
    else:
        # user indicated python version. Check if there is version compatibility with local python version and give user warnings as needed
        _check_local_python_dependency(dependency)

    yaml_template = {
        "name": name,
        "channels": channels,
        "dependencies": dependencies,
    }

    response_body = None
    start_time = time.time()
    with tempfile.NamedTemporaryFile(mode="w", suffix=".yml", delete=False) as definition_file:

        yaml.dump(yaml_template, definition_file, default_flow_style=False)

        client = get_client(settings)

        # Open a separate reader in binary mode per Requests doc
        try:
            with open(definition_file.name, "rb") as def_file_reader:
                data = {
                    "name": name,
                }
                # add the runtime_config only if base_image is provided
                if base_image:
                    image_config = EnvironmentRuntimeConfig(
                        image=BaseImageConfig(base_image=base_image, nvidia=nvidia)
                        if base_image
                        else None
                    )
                    data["runtime_config"] = image_config.model_dump_json(exclude_none=True)
                response = client.post(
                    "/api/v2/envs",
                    {
                        "files": {"definition": def_file_reader},
                        "data": data,
                    },
                )
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 400:
                print("Environment Already Exists.")
                return
            elif (
                e.response.status_code == 401 and e.response.json()["code"] == "auth/unauthorized"
            ):
                CovalentAPIKeyError(
                    message="A valid API key is required to create an environment.",
                    code=e.response.json()["code"],
                ).rich_print(level="error")
                return
            else:
                raise CovalentGenericAPIError(error=e) from e

        response_body = response.json()

    env_is_ready = False
    if wait:
        env_is_ready, env_response = check_env_is_ready(name, settings, wait=True, timeout=timeout)

    if env_is_ready:
        env_name = env_response["records"][0]["name"]
        env_status = env_response["records"][0]["status"]
        env_build_estimate = int(time.time() - start_time)
    else:
        env_name = response_body["name"]
        env_status = response_body["status"]
        env_build_estimate = response_body["estimated_time"]

    print(f"Name: {env_name}")
    print(f"Status: {env_status}")
    print(f"Estimated Time: {env_build_estimate} seconds")
    if notes:
        print("Notes:")
        for note in notes:
            print(f"\t{note}")

    print("Environment file contains:")
    print("==========================")
    print(yaml.dump(yaml_template, default_flow_style=False))


def _get_envs_filtered(**filters) -> Dict[str, Environment]:
    """
    Retrieves the envs from the Covalent Cloud server which match the filters.
    """

    client = get_client()

    try:

        response = client.get(
            "/api/v2/envs",
            request_options={
                "params": filters,
            },
        )

        response.raise_for_status()

        response_dict = response.json()

        envs = {}
        for env in response_dict["records"]:
            readable_env = {
                "id": env["id"],
                "name": env["name"],
                "created_at": env["created_at"],
                "status": env["status"],
                "definition_url": env["definition"],
            }
            envs[env["name"]] = Environment(**readable_env)

    except requests.exceptions.HTTPError:
        print("Error response: ", response.text)
        raise

    return envs


def get_envs(count: int = 10) -> Dict[str, Environment]:
    """
    Retrieves the last `count` created environments from the Covalent Cloud server which are ready.

    Returns:
        Dictionary of environments with their names as keys and their details as values.

    """

    envs = _get_envs_filtered(count=count, status="READY")
    return envs


def delete_env(env_name: str) -> None:
    """
    Deletes the environment with the given name from the Covalent Cloud server.

    Args:
        env_name: Name of the environment to be deleted.

    Returns:
        None

    """

    client = get_client()

    try:
        env = _get_envs_filtered(name=env_name)
        env_id = env[env_name].id

        response = client.delete(f"/api/v2/envs/{env_id}")
        response.raise_for_status()
        print(f"Environment {env_name} deleted successfully.")
    except requests.exceptions.HTTPError:
        print("Error response: ", response.text)
        raise



# Copyright 2023 Agnostiq Inc.

import requests
import requests.adapters
from furl import furl

from covalent_cloud.service_account_interface.auth_config_manager import AuthConfigManager
from covalent_cloud.shared.classes.settings import Settings, settings


class APIClient:

    api_key: str
    uri_components: furl
    auth_config: AuthConfigManager

    @property
    def host(self):
        return self.uri_components.host

    @property
    def port(self):
        return self.uri_components.port

    @property
    def scheme(self):
        return self.uri_components.scheme

    def __init__(
        self, host_uri, port=None, headers=None, settings=None, url_prefix="", max_retries=5
    ) -> None:
        if headers is None:
            headers = {}
        if settings is None:
            settings = Settings()

        self.uri_components = furl(host_uri)
        self.settings = settings
        self.headers = headers
        self.url_prefix = url_prefix

        self._adapter = requests.adapters.HTTPAdapter(max_retries=max_retries)

        if port:
            self.uri_components.set(port=port)

    def get_global_headers(self):
        auth_headers = AuthConfigManager.get_auth_request_headers(self.settings)
        return {**auth_headers, **self.headers}

    def get_request_options(self, request_options):
        request_options = request_options or {}
        provided_headers = request_options.get("headers") or {}
        options = {
            **request_options,
            "headers": {
                **self.get_global_headers(),
                **provided_headers,
            },
        }
        return options

    def prepare_request(self, endpoint, request_options=None):
        uri_components = self.uri_components.copy()
        uri_components.path = uri_components.path / self.url_prefix / endpoint
        uri = uri_components.url
        options = self.get_request_options(request_options)
        return uri, options

    def post(self, endpoint, request_options=None):
        uri, options = self.prepare_request(endpoint, request_options)
        with requests.Session() as session:
            session.mount("http://", self._adapter)
            session.mount("https://", self._adapter)
            res = session.post(uri, **options)

        res.raise_for_status()
        return res

    def get(self, endpoint, request_options=None):
        uri, options = self.prepare_request(endpoint, request_options)

        with requests.Session() as session:
            session.mount("http://", self._adapter)
            session.mount("https://", self._adapter)
            res = session.get(uri, **options)
        res.raise_for_status()
        return res

    def delete(self, endpoint, request_options=None):
        uri, options = self.prepare_request(endpoint, request_options)
        with requests.Session() as session:
            session.mount("http://", self._adapter)
            session.mount("https://", self._adapter)
            res = session.delete(uri, **options)
        res.raise_for_status()
        return res

    def put(self, endpoint, request_options=None):
        uri, options = self.prepare_request(endpoint, request_options)
        with requests.Session() as session:
            session.mount("http://", self._adapter)
            session.mount("https://", self._adapter)
            res = session.put(uri, **options)
        res.raise_for_status()
        return res


class DispatcherAPI(APIClient):
    def __init__(self, headers=None, settings: Settings = settings) -> None:
        if headers is None:
            headers = {}

        super().__init__(
            host_uri=settings.dispatcher_uri,
            port=settings.dispatcher_port,
            headers=headers,
            settings=settings,
        )


class DeploymentAPI(APIClient):
    def __init__(self, headers=None, settings: Settings = settings) -> None:
        if headers is None:
            headers = {}

        # To test local run:
        # super().__init__(
        #     host_uri="http://127.0.0.1",
        #     port=8080,
        #     headers=headers,
        #     settings=settings,
        #     url_prefix="/api/v0",
        # )

        # This uses the same uri and port as the dispatcher but with url_prefix added
        super().__init__(
            host_uri=settings.dispatcher_uri,
            port=settings.dispatcher_port,
            headers=headers,
            settings=settings,
            url_prefix="/fn/api/v0",
        )



# Copyright 2023 Agnostiq Inc.

"""Module for Covalent Cloud dispatching and related functionalities."""


from copy import deepcopy
from functools import wraps
from typing import Any, Callable, Dict, List, Union

import covalent as ct
from covalent._results_manager.result import Result
from covalent._workflow.lattice import Lattice

from covalent_cloud.service_account_interface.client import get_client
from covalent_cloud.shared.classes.exceptions import handle_error
from covalent_cloud.shared.schemas.volume import Volume

from ..shared.classes.settings import Settings, settings
from . import results_manager as rm
from .dispatch_info import DispatchInfo
from .helpers import (
    add_dispatch_info,
    fast_redispatch,
    register,
    start,
    track_redispatch,
    validate_executors,
)

API_KEY = "fake"  # pragma: allowlist secret


def cancel(dispatch_id: str, task_ids: List[int] = None, settings: Settings = settings) -> None:
    """Cancel a running dispatch.

    Args:
        dispatch_id: The dispatch id of the dispatch to be cancelled.
        task_ids: Optional, list of task ids to cancel within the workflow
        settings: The settings object to use. If None, the default settings will be used.

    """
    task_ids = [task_ids] if isinstance(task_ids, int) else (task_ids or [])

    api_client = get_client(settings=settings)
    try:
        response = api_client.put(
            f"api/v2/lattices/{dispatch_id}/cancel",
            request_options={"json": {"task_ids": task_ids}},
        )
        print(response.json().get("message"))
    except Exception as e:
        print(f"Error cancelling dispatch {dispatch_id}")


def dispatch(
    orig_lattice: Lattice,
    settings: Settings = settings,
    volume: Union[Volume, None] = None,
    disable_run: bool = False,
    dispatch_info: Union[DispatchInfo, Dict[str, Any], None] = None,
    tags: Union[List[str], None] = None,
) -> Callable:
    """
    Dispatches a Covalent workflow to the Covalent Cloud and returns the assigned dispatch ID.
    The dispatch function takes a Covalent workflow, also called a lattice, and sends it to the Covalent Cloud Server for execution. Once dispatched, the workflow runs on a the cloud and can be monitored using the dispatch ID in the application. The dispatch function returns a wrapper function that takes the inputs of the workflow as arguments. This wrapper function can be called to execute the workflow.

    Args:
        orig_lattice: The Covalent workflow to send to the cloud.
        settings: The settings object to use. If None, the default settings will be used.
        volume: [optional] Volume instance
        dispatch_info: [optional] DispatchInfo object to attach to a workflow - this will take priority over `tags` and other `DispatchInfo` parameters if both are provided.
        tags: [optional] List of tags to attach to a workflow - will be converted to a DispatchInfo object.

    Returns:
        A wrapper function which takes the inputs of the workflow as arguments.

    Examples:

        .. highlight:: python
        .. code-block:: python

            # define a simple lattice
            import covalent_cloud as cc
            import covalent as ct

            # create volume [optional]
            volume = cc.volume("/myvolume")

            @ct.lattice
            def my_lattice(a: int, b: int) -> int:
                return a + b

            # dispatch the lattice and get the assigned dispatch ID
            dispatch_id = cc.dispatch(my_lattice, volume=volume)(2, 3)
    """

    @wraps(orig_lattice)
    def wrapper(*args, **kwargs) -> str:
        """
        Send the lattice to the dispatcher server and return
        the assigned dispatch id.

        Args:
            *args: The inputs of the workflow.
            **kwargs: The keyword arguments of the workflow.

        Returns:
            The dispatch id of the workflow.

        """

        try:

            lattice = deepcopy(orig_lattice)

            # Enabling task_packing for the build_graph call as it isn't
            # supported yet by OS covalent.
            old_task_packing = ct.get_config("sdk.task_packing")
            ct.set_config("sdk.task_packing", "true")
            lattice.build_graph(*args, **kwargs)
            ct.set_config("sdk.task_packing", old_task_packing)

            # Check that only CloudExecutors are specified.
            if settings.validate_executors and not validate_executors(lattice):
                raise ValueError("One or more electrons have invalid executors.")

            # TODO: Update register endpoint to take volume_id as a param and associate EFS to job_def
            dispatch_id = register(lattice, volume=volume, settings=settings)(*args, **kwargs)

            # The dispatch id is ALSO returned by the start function - both the register and the start function's dispatch_ids are the same
            if not disable_run:
                dispatch_id = start(dispatch_id, settings)
            else:
                return dispatch_id

            # Making sure the dispatch_info is the one received from the parent function
            nonlocal dispatch_info

            # The `DispatchInfo` object takes priority over `tags` and other `DispatchInfo` parameters if both are provided.
            if dispatch_info:
                if isinstance(dispatch_info, dict):
                    dispatch_info = DispatchInfo(**dispatch_info)
                add_dispatch_info(dispatch_id, dispatch_info, settings)
            elif tags:
                add_dispatch_info(dispatch_id, DispatchInfo(tags=tags), settings)

            return dispatch_id

        except Exception as e:
            handle_error(e)

    return wrapper


def redispatch(
    dispatch_id: str,
    rebuild: bool = False,
    settings: Settings = settings,
) -> Callable[..., str]:
    """
    Re-dispatches a Covalent workflow to the Covalent Cloud and returns the assigned dispatch ID.

    Args:
        dispatch_id: The dispatch ID of the workflow to re-dispatch.
        rebuild: Whether to rebuild the workflow locally before re-dispatching. Defaults to False.
        settings: The settings object to use. If None, the default settings will be used.

    Returns:
        Callable redispatch workflow function.

    Examples:

    .. highlight:: python
    .. code-block:: python

        import time
        import covalent as ct
        import covalent_cloud as cc

        cc.save_api_key("your-api-key")

        default_exec = cc.CloudExecutor()

        @ct.electron(executor=default_exec)
        def add(a: int, b: int) -> int:
            return a + b

        @ct.lattice(executor=default_exec, workflow_executor=default_exec)
        def my_lattice(a: int, b: int) -> int:
            return add(a, b)

        # dispatch the lattice and get the assigned dispatch ID
        dispatch_id_0 = cc.dispatch(my_lattice)(2, 3)
        # wait for the result...
        result_0 = cc.get_result(dispatch_id_0, wait=True)

        # re-dispatch the lattice with new inputs without rebuilding locally
        redispatch_id_1 = cc.redispatch(dispatch_id_0)([3, 4], {})
        time.sleep(60) # wait for the new dispatch to register and start
        # wait for the new result...
        result_1 = cc.get_result(redispatch_id_1, wait=True)

        # re-dispatch the lattice with new inputs, rebuilding locally
        redispatch_id_2 = cc.redispatch(dispatch_id_0, rebuild=True)([5, 6], {})
        # wait for another new result...
        result_2 = cc.get_result(redispatch_id_2, wait=True)

    """

    def func(*input_args: Any, **input_kwargs: Any) -> str:
        """Redispatch workflow function.

        Args:
            input_args: The positional arguments of the workflow.
            input_kwargs: The keyword arguments of the workflow.

        Returns:
            The dispatch ID of the re-dispatched workflow.

        """
        message = ""
        if rebuild:
            response = rebuild_redispatch(dispatch_id, input_args, input_kwargs, settings)
            redispatch_id = response
            message = f"Rebuilt and re-dispatched workflow, original dispatch_id: {dispatch_id}, new dispatch_id: {redispatch_id}"
        else:
            response = fast_redispatch(dispatch_id, input_args, input_kwargs, settings)
            redispatch_id = response
            message = f"Re-dispatched workflow, original dispatch_id: {dispatch_id}, new dispatch_id: {redispatch_id}"

        print(message)
        return redispatch_id

    return func


def get_result(
    dispatch_id: str,
    wait: bool = False,
    settings: Settings = settings,
    *,
    status_only: bool = False,
) -> Result:
    """
    Gets the result of a Covalent workflow that has been dispatched to the cloud.

    This function retrieves the result of a dispatched Covalent workflow that has been executed on the cloud. The function takes the dispatch ID of the workflow and retrieves the results from the cloud. The result is returned as a `Result` object that contains the status of the workflow, the final result of the lattice, and any error messages that may have occurred during execution.

    Args:
        dispatch_id: The dispatch ID assigned to the workflow.
        wait: Controls how long the function waits for the server to return a result. If False, the function will not wait and will return the current status of the workflow. If True, the function will wait for the result to finish and keep retrying until the result is available.
        status_only: If True, only retrieves the status of the workflow and not the full result.
        settings: The Covalent settings to use for the request.

    Returns:
        A `Result` object that contains the status of the workflow, the final result of the lattice, and any error messages that may have occurred during execution.

    Examples:

        .. highlight:: python
        .. code-block:: python

            # define a simple lattice
            import covalent_cloud as cc
            import covalent as ct

            @ct.lattice
            def my_lattice(a: int, b: int) -> int:
                return a + b

            # dispatch the lattice and get the assigned dispatch ID
            dispatch_id = cc.dispatch(my_lattice)(2, 3)

            # get the result of the dispatched lattice
            result = cc.get_result(dispatch_id, wait=True)

            # print the final result of the lattice
            print(result.result)
    """

    return rm.get_result(
        dispatch_id=dispatch_id,
        wait=wait,
        settings=settings,
        status_only=status_only,
    )


def rebuild_redispatch(
    dispatch_id: str,
    input_args: list,
    input_kwargs: dict,
    settings: Settings = settings,
) -> Callable:
    """
    Re-dispatches a Covalent workflow to the Covalent Cloud and returns the assigned dispatch ID.

    Args:
        dispatch_id: The dispatch ID of the workflow to re-dispatch.
        input_args: The positional arguments of the workflow.
        input_kwargs: The keyword arguments of the workflow.
        settings: The settings object to use. If None, the default settings will be used.

    Returns:
        The dispatch ID of the re-dispatched workflow.

    """

    try:
        # loading original dispatch
        original_dispatch = get_result(dispatch_id, wait=True, settings=settings)
        original_dispatch.result.load()

        # load lattice
        original_dispatch.lattice.workflow_function.load()
        original_dispatch.lattice.metadata["hooks"].load()

        # build lattice
        redispatch_lattice = ct.lattice(original_dispatch.lattice.workflow_function.value)
        redispatch_lattice.metadata = original_dispatch.lattice.metadata.copy()
        redispatch_lattice.metadata["hooks"] = original_dispatch.lattice.metadata[
            "hooks"
        ].value.copy()

        redispatch_id = dispatch(redispatch_lattice, settings=settings)(
            *input_args, **input_kwargs
        )

        # Update redispatch with original dispatch id
        update_response = track_redispatch(redispatch_id, dispatch_id, settings)
        if (
            update_response["dispatch_id"] != redispatch_id
            or update_response["original_dispatch_id"] != dispatch_id
        ):
            raise Exception(
                f"Failed to update redispatch: {redispatch_id} with original dispatch id: {dispatch_id}"
            )

    except Exception as e:
        raise e

    return redispatch_id



# Copyright 2024 Agnostiq Inc.
"""Test behavior while waiting for deployments"""

import time
import unittest
from unittest.mock import patch

import covalent_cloud as cc
from covalent_cloud.function_serve.common import ACTIVE_DEPLOYMENT_POLL_INTERVAL as POLL_FREQ
from covalent_cloud.function_serve.common import ServiceStatus
from covalent_cloud.function_serve.models import Deployment
from covalent_cloud.shared.classes.api import APIClient


class DeploymentResponse:
    """Response from which a deployment can be created."""

    def __init__(self, deployment: Deployment):
        self.deployment = deployment

    def json(self) -> dict:
        return {
            "id": self.deployment.function_id,
            "description": self.deployment.description,
            "invoke_url": self.deployment.address,
            "title": self.deployment.name,
            "endpoints": [],
            "tags": self.deployment.tags,
            "status": self.deployment.status,
            "inference_keys": [],
            "auth": False,
            "error": None,
            "stderr": None,
        }


class TestGetDeploymentWait(unittest.TestCase):
    """Test waiting behavior."""

    def _get_deployment_kwargs(self, **kwargs) -> dict:
        return {
            "function_id": "1234567890",
            "address": "http://fake-address.com",
            "name": "Fake Service for Testing",
            "description": "A fake service for testing purposes.",
            "routes": [],
            "status": ServiceStatus.ACTIVE,
            "tags": ["fake-tag-1", "fake-tag-2"],
            **kwargs,
        }

    def setUp(self) -> None:
        self.deployment = Deployment(**self._get_deployment_kwargs())

    def test_get_deployment_without_wait(self):
        """Test without waiting for deployment to be active."""

        def _mock_deployment_client_get(*_, **__):
            return DeploymentResponse(_mock_deployment_client_get.deployment)

        for status in ServiceStatus:
            dep = Deployment(**self._get_deployment_kwargs(status=status))
            _mock_deployment_client_get.deployment = dep
            with patch.object(APIClient, "get", new=_mock_deployment_client_get):
                t0 = time.monotonic()
                cc.get_deployment(dep, wait=False)
                tf = time.monotonic()
                assert (tf - t0) < POLL_FREQ, f"Status {status!s} should not wait if wait=False"

    def test_get_deployment_invalid_wait(self):
        """Test without waiting for deployment to be active."""

        def _mock_deployment_client_get(*_, **__):
            return DeploymentResponse(self.deployment)

        with patch.object(APIClient, "get", new=_mock_deployment_client_get):

            with self.assertRaises(ValueError):
                cc.get_deployment(self.deployment, wait="a string is invalid")

            with self.assertRaises(ValueError):
                cc.get_deployment(self.deployment, wait=["nor", "a", "list"])

            with self.assertRaises(ValueError):
                cc.get_deployment(self.deployment, wait={"not": "a", "dict": "either"})

            with self.assertRaises(ValueError):
                cc.get_deployment(self.deployment, wait=-1)  # negatives too, not allowed



# Copyright 2023 Agnostiq Inc.

import responses

from covalent_cloud.shared.classes.exceptions import CovalentSDKError
from covalent_cloud.shared.classes.helpers import check_env_is_ready
from covalent_cloud.shared.classes.settings import Settings


class TestHelpersShould:
    @responses.activate
    def test_check_env_is_ready_returns_true_when_ready(self):
        settings = Settings()
        settings.dispatcher_uri = "https://api.test.covalent.xyz"
        responses.add(
            responses.GET,
            "https://api.test.covalent.xyz/api/v2/envs/?name=env_name",
            status=200,
            json={"records": [{"status": "READY"}]},
        )

        env_is_ready, env_response = check_env_is_ready("env_name", settings)

        assert env_is_ready == True  # noqa: E712

    @responses.activate
    def test_check_env_is_ready_throws_exception_when_dne(self):
        settings = Settings()
        settings.dispatcher_uri = "https://api.test.covalent.xyz"
        responses.add(
            responses.GET,
            "https://api.test.covalent.xyz/api/v2/envs/?name=env_name",
            status=200,
            json={"records": []},
        )
        try:
            env_is_ready, env_response = check_env_is_ready("env_name", settings)
        except CovalentSDKError as e:
            assert 'Environment "env_name" does not exist.' in str(e)

    @responses.activate
    def test_check_env_is_ready_throws_exception_when_not_ready(self):
        settings = Settings()
        settings.dispatcher_uri = "https://api.test.covalent.xyz"
        responses.add(
            responses.GET,
            "https://api.test.covalent.xyz/api/v2/envs/?name=env_name",
            status=200,
            json={"records": [{"status": "CREATING"}]},
        )
        try:
            env_is_ready, env_response = check_env_is_ready("env_name", settings)
        except CovalentSDKError as e:
            assert 'Environment "env_name" is not ready.' in str(e)



# Copyright 2024 Agnostiq Inc.
"""Tests around deploying services"""
from contextlib import contextmanager
from unittest.mock import Mock, patch

import pytest

import covalent_cloud as cc
from covalent_cloud.function_serve.assets import AssetsMediator
from covalent_cloud.function_serve.deployment import deploy
from covalent_cloud.function_serve.models import Deployment
from covalent_cloud.function_serve.service_class import FunctionService
from covalent_cloud.shared.classes.api import DeploymentAPI


def sample_func(x, /, y, *args, foo="default", k, bar="kw_default", **kwargs):
    """
    Sample service to test deployment
    """
    return {"x": x, "y": y, "args": args, "foo": foo, "k": k, "bar": bar, "kwargs": kwargs}


sample_service = cc.service(
    sample_func, executor=Mock(memory=20_480)  # will break real deployment attempt
)


@contextmanager
def mock_deploy_to_cloud():
    """Patch everything inside the deploy wrapper"""
    with patch.object(DeploymentAPI, "post"), patch.object(
        FunctionService, "get_model"
    ), patch.object(AssetsMediator, "hydrate_assets_from_model"), patch.object(
        AssetsMediator, "upload_all"
    ), patch.object(
        Deployment, "from_function_record"
    ), patch.object(
        Deployment, "attach_route_methods"
    ):

        yield


@pytest.mark.parametrize(
    "args, kwargs",
    [
        # x, y, ... (minimum args & kwargs)
        ((1, 2), {"k": 99.9}),
        # x, y, ... (everything except bar)
        ((1, 2), {"k": 99.9, "foo": "dog", "extra_kw1": "value-1", "extra_kw2": "value-2"}),
        # x, y, foo="dog", ...
        ((1, 2, "dog"), {"k": 99.9, "bar": "cat", "extra_kw1": "value-1", "extra_kw2": "value-2"}),
        # x, y, foo, *args
        (
            (1, 2, "dog", 4, 5, 6),
            {"k": 99.9, "bar": "cat", "extra_kw1": "value-1", "extra_kw2": "value-2"},
        ),
    ],
)
def test_deploy_signature_valid(args, kwargs):
    """All calls here should pass the signature check."""

    sample_func(*args, **kwargs)  # NOTE: test validation

    with mock_deploy_to_cloud():
        deploy(sample_service)(*args, **kwargs)


@pytest.mark.parametrize(
    "args, kwargs",
    [
        # x, y, ... (missing k)
        ((1, 2), {"foo": "dog"}),
        # x, y, ... (missing k)
        ((1, 2), {"foo": "dog", "extra_kw1": "value-1", "extra_kw2": "value-2"}),
        # x ... (missing y)
        ((1,), {"k": 99.9, "bar": "cat", "extra_kw1": "value-1", "extra_kw2": "value-2"}),
        # x, y, foo, *args
        ((1, 2, "dog"), {"bar": "cat", "extra_kw1": "value-1", "extra_kw2": "value-2"}),
        # totally wrong
        ((exec,), {"order": 66}),
        # empty
        ((), {}),
        # no kwargs  (missing k)
        ((1, 2, "dog", 4, 5, 6), {}),
        # x not positional-only
        ((2,), {"x": 1, "k": 99.9}),
    ],
)
def test_deploy_signature_invalid(args, kwargs):
    """All calls here should fail the signature check."""
    with pytest.raises(TypeError):
        sample_func(*args, **kwargs)  # NOTE: test validation

    with mock_deploy_to_cloud():
        with pytest.raises(TypeError):
            deploy(sample_service)(*args, **kwargs)



# Copyright 2023 Agnostiq Inc.

import site
import sys

from setuptools import find_packages, setup

site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

with open("VERSION") as f:
    version = f.read().strip()

with open("requirements.txt") as f:
    required = f.read().splitlines()

setup_info = {
    "name": "covalent_cloud",
    "packages": find_packages(exclude=["tests"]),
    "version": version,
    "maintainer": "Agnostiq",
    "url": "https://github.com/AgnostiqHQ/covalent-cloud-sdk",
    "download_url": f"https://github.com/AgnostiqHQ/covalent-cloud-sdk/archive/v{version}.tar.gz",
    "license": "Apache License 2.0",
    "author": "Agnostiq",
    "author_email": "support@agnostiq.ai",
    "description": "Covalent Cloud SDK",
    "long_description": open("README.md").read(),
    "long_description_content_type": "text/markdown",
    "include_package_data": True,
    "install_requires": required,
    "python_requires": ">=3.8",
    "classifiers": [
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: Other/Proprietary License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Adaptive Technologies",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Software Development",
        "Topic :: System :: Distributed Computing",
    ],
}

if __name__ == "__main__":
    setup(**setup_info)



# Copyright 2024 Agnostiq Inc.

from enum import Enum
from typing import Any, Dict, List, Optional, Union

import requests
from pydantic import BaseModel, ConfigDict


class BaseImageConfig(BaseModel):
    """
    Represents the base Docker image configuration for an environment.

    Attributes:
        base_image: The base Docker image for the environment, formatted as `<registry>/<image>:<tag>`,
                    where `:tag` is optional. This should point to a publicly accessible or private image.

        username_credentials_reference_id: Optional(str); The ID of the username secret uploaded to the Covalent Cloud via
                                            the `store_secret` function, if the Docker image is stored in a private.

        password_credentials_reference_id: Optional(str); The ID of the password uploaded to the Covalent Cloud via the
                                        `store_secret` function, if the Docker image is stored in a private registry.
    """

    base_image: str
    # Temporary: handle Nvidia images as a special case until we
    # refactor the general base_image pipeline
    nvidia: bool = False
    username_credentials_reference_id: Optional[str] = None
    password_credentials_reference_id: Optional[str] = None
    model_config = ConfigDict(from_attributes=True)


class PackageProvider(Enum):
    """
    Enum representing the package provider for runtime packages to be installed in the environment.

    Values:
        SYSTEM: the OS default package manager (e.g., APT for Ubuntu, yum for CentOS, etc.).
        WGET: wget to download packages from a given URL.
        GIT: git to clone repositories as packages.
    """

    SYSTEM = "SYSTEM"
    WGET = "WGET"
    GIT = "GIT"


class RuntimePackage(BaseModel):
    """
    Defines a runtime package to be installed in the environment.

    Attributes:
        name: The name of the package. For SYSTEM providers, this is the package name. For WGET and GIT,
              this should be the URL of the artifact or repository to fetch.
        provider Optional(PackageProvider): The provider for this package. Defaults to SYSTEM if not specified.
    """

    name: str
    provider: Optional[PackageProvider] = PackageProvider.SYSTEM
    model_config = ConfigDict(from_attributes=True)


class EnvironmentRuntimeConfig(BaseModel):
    """
    Configuration for the runtime environment, including the base image and any additional packages.

    Attributes:
        image: Optional(BaseImageConfig); Configuration for the base Docker image of the environment.
    """

    image: Optional[BaseImageConfig] = None
    model_config = ConfigDict(from_attributes=True)


class Environment(BaseModel):
    """
    Basic representation of a user created environment in Covalent Cloud.

    Attributes:
        id: The unique identifier of the environment.
        name: The name of the environment.
        created_at: The timestamp when the environment was created.
        status: The current status of the environment.
        definition: A dictionary containing the environment definition.
    """

    id: str
    name: str
    created_at: str
    status: str
    definition_url: str

    # Will be populated when the environment definition is loaded
    conda: Optional[List[str]] = None  # List of conda packages
    pip: Optional[List[str]] = None  # List of pip packages

    def __str__(self) -> str:
        return f"Environment({self.name})"

    def __repr__(self) -> str:
        return f"Environment({self.name})"

    def to_dict(self) -> Dict[str, Any]:
        return self.model_dump()

    def _parse_definition(self, text: str) -> Dict[str, Union[str, List[str]]]:
        """
        Parse the environment definition from the raw string to a dictionary.
        """

        conda = []
        pip = []

        lines = text.split("\n")
        # Only get the lines that start with "-" or "  -" i.e. the dependencies
        lines = [line for line in lines if line.startswith("-") or line.startswith("  -")]
        i = 0

        # Parse up to the first "pip:" line to get all the conda packages
        while not lines[i].endswith("pip:"):
            if lines[i].startswith("-"):
                package = lines[i].strip("-").strip()
                conda.append(package)
            i += 1

        # Parse the pip packages after the "pip:" line
        i += 1
        while i < len(lines) and not lines[i].startswith("-"):
            if lines[i].startswith("  -"):
                package = lines[i].strip("  -").strip()
                pip.append(package)
            i += 1

        return conda, pip

    def load_definition(self) -> None:
        """
        Load the environment definition from presigned S3 URL.
        """

        # In case the definition is already loaded in any of the attributes
        if self.conda is not None or self.pip is not None:
            return

        response = requests.get(self.definition_url)
        response.raise_for_status()
        self.conda, self.pip = self._parse_definition(response.text)
